from __future__ import annotations

import configparser
import dataclasses
import datetime
import os
from logging import getLogger
from os import PathLike
from pathlib import Path
from typing import TypeVar

from cachestore.formatters import Formatter, PickleFormatter
from cachestore.hashers import Hasher, PickleHasher
from cachestore.storages import LocalStorage, Storage
from cachestore.util import safe_import_object

DISABLE_CACHE = os.environ.get("CACHESTORE_DISABLE", "0").lower() in ("1", "true")

logger = getLogger(__name__)


T = TypeVar("T")


@dataclasses.dataclass
class CacheSettings:
    storage: Storage = dataclasses.field(default_factory=lambda: LocalStorage())
    formatter: Formatter = dataclasses.field(default_factory=lambda: PickleFormatter())
    hasher: Hasher = dataclasses.field(default_factory=lambda: PickleHasher())
    disable: bool = DISABLE_CACHE


@dataclasses.dataclass
class FunctionSettings:
    ignore: set[str] = dataclasses.field(default_factory=set)
    expire: int | datetime.timedelta | datetime.date | datetime.datetime | None = None
    formatter: Formatter | None = None
    disable: bool | None = None

    @property
    def expired_at(self) -> datetime.datetime | None:
        if isinstance(self.expire, int):
            expired_at = datetime.datetime.now() + datetime.timedelta(days=self.expire)
        elif isinstance(self.expire, datetime.timedelta):
            expired_at = datetime.datetime.now() + self.expire
        elif isinstance(self.expire, datetime.datetime):
            expired_at = self.expire
        elif isinstance(self.expire, datetime.date):
            expired_at = datetime.datetime(
                year=self.expire.year,
                month=self.expire.month,
                day=self.expire.day,
            )
        else:
            expired_at = None
        return expired_at


class Config:
    def __init__(self, filename: str | PathLike | None = None) -> None:
        self._parser = configparser.ConfigParser()
        self._cache_settings: dict[str, CacheSettings] = {}
        self._function_settings: dict[str, FunctionSettings] = {}

        filenames: list[Path] = []
        if filename:
            filenames.append(Path(filename))
        if filename := os.environ.get("CACHESTORE_CONFIG_PATH"):
            filenames.append(Path(filename))
        filenames.append(Path.cwd() / "cachestore.ini")

        for filename in filenames:
            if filename.exists():
                logger.info("Load config from %s", filename)
                self._parser.read(filename)

        # load cache instances
        for section in self._parser.sections():
            if self._is_function_section(section):
                self._function_settings[section] = self._load_function_settings(self._parser[section])
            else:
                self._cache_settings[section] = self._load_cache_settings(self._parser[section])

    def cache_settings(self, name: str) -> CacheSettings:
        if name not in self._cache_settings:
            self._cache_settings[name] = CacheSettings()
        return self._cache_settings[name]

    def function_settings(self, name: str) -> FunctionSettings:
        if name not in self._function_settings:
            self._function_settings[name] = FunctionSettings()
        return self._function_settings[name]

    def _is_function_section(self, name: str) -> bool:
        return " " in name

    def _load_cache_settings(self, config: configparser.SectionProxy) -> CacheSettings:
        settings = CacheSettings()
        if "storage" in config:
            storagecls = safe_import_object(config["storage"])
            assert issubclass(storagecls, Storage)
            settings.storage = storagecls.from_config(config)
        if "formatter" in config:
            formattercls = safe_import_object(config["formatter"])
            assert issubclass(formattercls, Formatter)
            settings.formatter = formattercls.from_config(config)
        if "hasher" in config:
            hashercls = safe_import_object(config["hasher"])
            assert issubclass(hashercls, Hasher)
            settings.hasher = hashercls.from_config(config)
        settings.disable = config.getboolean("disable", settings.disable)
        return settings

    def _load_function_settings(self, config: configparser.SectionProxy) -> FunctionSettings:
        settings = FunctionSettings()
        if "ignore" in config:
            settings.ignore = set(x.strip() for x in config["ignore"].split(","))
        if "expire" in config:
            settings.expire = datetime.datetime.fromisoformat(config["expire"])
        if "formatter" in config:
            formattercls = safe_import_object(config["formatter"])
            assert issubclass(formattercls, Formatter)
            settings.formatter = formattercls.from_config(config)
        if "disable" in config:
            settings.disable = config.getboolean("disable", settings.disable)
        return settings



from __future__ import annotations

import importlib
import inspect
import pkgutil
import string
import sys
from types import ModuleType
from typing import Any


def b62encode(data: bytes) -> str:
    num = int.from_bytes(data, "big")
    characters = string.digits + string.ascii_letters

    encoded = ""
    base: int = len(characters)

    if num < 0:
        return ""

    while num >= base:
        mod = num % base
        num //= base
        encoded = characters[mod] + encoded

    if num > 0:
        encoded = characters[num] + encoded

    return encoded


def import_submodules(package_name: str) -> None:
    importlib.invalidate_caches()

    sys.path.append(".")

    # Import at top level
    module = safe_import_module(package_name)
    path = getattr(module, "__path__", [])
    path_string = "" if not path else path[0]

    for module_finder, name, _ in pkgutil.walk_packages(path):
        if path_string and getattr(module_finder, "path") != path_string:  # noqa: B009
            continue
        subpackage = f"{package_name}.{name}"
        import_submodules(subpackage)


def import_modules(module_names: list[str]) -> None:
    for module_name in module_names:
        import_submodules(module_name)


def safe_import_module(modulename: str) -> ModuleType:
    for name, module in list(sys.modules.items()):
        try:
            if name == modulename or (
                hasattr(module, "__file__") and modulename == inspect.getmodulename(inspect.getabsfile(module))
            ):
                return module
        except AttributeError:
            pass
    return importlib.import_module(modulename)


def safe_import_object(path: str) -> Any:
    if "." not in path and ":" not in path:
        raise ValueError("Cache name must be formatted as path.to.module:name.")
    if ":" in path:
        modulename, objname = path.rsplit(":", 1)
    else:
        modulename, objname = path.rsplit(".", 1)
    module = safe_import_module(modulename)
    return getattr(module, objname)


def find_variable_path(obj: Any) -> str | None:
    for modulename, module in list(sys.modules.items()):
        if hasattr(module, "__file__"):
            modulename = inspect.getmodulename(inspect.getabsfile(module)) or modulename
        try:
            for varname, value in module.__dict__.items():
                if value is obj:
                    return f"{modulename}:{varname}"
        except AttributeError:
            pass
    return None



from __future__ import annotations

import os
from configparser import SectionProxy
from contextlib import contextmanager
from os import PathLike
from pathlib import Path
from typing import IO, Any, Callable, ContextManager, Iterator, Type, TypeVar

from cachestore.common import FileLock
from cachestore.storages.storage import Storage
from cachestore.util import safe_import_object

DEFAULT_ROOT_DIR = ".cachestore"

Self = TypeVar("Self", bound="LocalStorage")


class LocalStorage(Storage):
    def __init__(
        self,
        root: str | PathLike | None = None,
        openfn: Callable[..., IO[Any]] | Callable[..., ContextManager[IO[Any]]] | None = None,
    ) -> None:
        self._root = Path(root or DEFAULT_ROOT_DIR).absolute()
        self._openfn = openfn or open

    def __str__(self) -> str:
        return f"LocalStorage(root={self._root.relative_to(Path.cwd())})"

    def __repr__(self) -> str:
        return f"LocalStorage(root={self._root.relative_to(Path.cwd())})"

    @contextmanager
    def open(self, key: str, mode: str) -> Iterator[IO[Any]]:
        filename = self._root / key
        lockfile = filename.parent / (filename.name + ".lock")
        filename.parent.mkdir(parents=True, exist_ok=True)
        with FileLock(lockfile):
            try:
                with self._openfn(filename, mode) as fp:
                    yield fp
            except (Exception, KeyboardInterrupt):
                os.remove(filename)
                raise
            finally:
                os.remove(lockfile)

    def remove(self, key: str) -> None:
        filename = self._root / key
        os.remove(filename)

    def exists(self, key: str) -> bool:
        return (self._root / key).exists()

    def all(self) -> Iterator[str]:
        for filename in self._root.glob("*"):
            yield filename.name

    def filter(self, prefix: str) -> Iterator[str]:
        for filename in self._root.glob(f"{prefix}*"):
            yield filename.name

    @classmethod
    def from_config(cls: Type[Self], config: SectionProxy) -> Self:
        root = config.get("storage.root")

        if "storage.openfn" in config:
            openfn = safe_import_object(config["storage.openfn"])
        else:
            openfn = None

        return cls(root=root, openfn=openfn)



import hashlib
import io
import pickle
from configparser import SectionProxy
from typing import Any, Type, TypeVar

from cachestore.hashers.hasher import Hasher
from cachestore.util import b62encode

Self = TypeVar("Self", bound="Hasher")


class PickleHasher(Hasher):
    def __call__(self, obj: Any) -> str:
        m = hashlib.blake2b()
        with io.BytesIO() as buf:
            pickle.dump(obj, buf)
            m.update(buf.getbuffer())
            return b62encode(m.digest())

    @classmethod
    def from_config(cls: Type[Self], config: SectionProxy) -> Self:
        return cls()



from __future__ import annotations

import fcntl
from pathlib import Path
from types import TracebackType
from typing import IO, Any, Type


class FileLock:
    def __init__(self, lockfile: str | Path) -> None:
        self._file_path = lockfile
        self._lockfile: IO[Any] | None = None

    def acquire(self) -> None:
        if self._lockfile is None:
            self._lockfile = open(self._file_path, "w")
            fcntl.flock(self._lockfile, fcntl.LOCK_EX)

    def release(self) -> None:
        if self._lockfile is not None:
            fcntl.flock(self._lockfile, fcntl.LOCK_UN)
            self._lockfile.close()
            self._lockfile = None

    def __enter__(self) -> FileLock:
        self.acquire()
        return self

    def __exit__(
        self,
        exc_type: Type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: TracebackType | None,
    ) -> bool:
        self.release()
        return exc_type is None and exc_value is None and traceback is None



import argparse
import sys

from cachestore.cache import Cache
from cachestore.commands.subcommand import Subcommand
from cachestore.common import Selector
from cachestore.util import import_modules, safe_import_object


@Subcommand.register("remove")
class RemoveCommand(Subcommand):
    """remove caches"""

    def setup(self) -> None:
        self.parser.add_argument("cache", help="cache name")
        self.parser.add_argument(
            "-f",
            "--function",
            action="append",
            default=[],
            help="function names to remove cache",
        )
        self.parser.add_argument(
            "-a",
            "--all",
            action="store_true",
            help="remove all caches of the specified cache",
        )
        self.parser.add_argument(
            "--include-package",
            action="append",
            default=[],
            help="additinoal packages to include",
        )

    def run(self, args: argparse.Namespace) -> None:
        if args.include_package:
            import_modules(args.include_package)

        cache = Cache.by_name(args.cache)
        if cache is None:
            cache = safe_import_object(args.cache)

        if cache is None:
            print(f"Given cache name is not found: {args.cache}", file=sys.stderr)
            sys.exit(1)

        funcinfos = {funcinfo.name: funcinfo for funcinfo in cache.funcinfos() if cache.exists(funcinfo)}
        if args.all:
            funcnames = set(funcinfos.keys())
        elif args.function:
            funcnames = set(args.function)
        else:
            funcnames = set()
            candidates = [name for name, info in funcinfos.items()]
            if candidates:
                selected_funcname = Selector()(candidates)
                funcnames = set() if selected_funcname is None else {selected_funcname}

        if funcnames:
            for funcname in funcnames:
                if args.all or funcname in funcinfos:
                    print(f"remove: {funcname}")
                    cache.remove(funcinfos[funcname])
                else:
                    print(f"skip  : {funcname}")
        else:
            print("No caches to remove.")



from __future__ import annotations

from importlib.metadata import version

from cachestore.cache import Cache  # noqa: F401
from cachestore.formatters import Formatter, PickleFormatter  # noqa: F401
from cachestore.hashers import Hasher, PickleHasher  # noqa: F401
from cachestore.storages import LocalStorage, Storage  # noqa: F401

__version__ = version("cachestore")
__all__ = [
    "__version__",
    "Cache",
    "Formatter",
    "PickleFormatter",
    "Hasher",
    "PickleHasher",
    "Storage",
    "LocalStorage",
]



from __future__ import annotations

import argparse
import re
from collections import defaultdict
from typing import Callable, ClassVar, Dict, NamedTuple, Optional, Type, TypeVar, cast

Subclass = TypeVar("Subclass", bound="Subcommand")
Registry = Dict[Type["Subcommand"], Dict[str, Type["Subcommand"]]]


class SubcommandInfo(NamedTuple):
    name: str
    usage: str | None = None
    description: str | None = None
    epilog: str | None = None


class Subcommand:
    _registry: ClassVar[Registry] = defaultdict(dict)
    _func_key: ClassVar[str] = "__func"
    _cmd_info: ClassVar[SubcommandInfo]

    @classmethod
    def register(
        cls,
        name: str | None = None,
        usage: str | None = None,
        description: str | None = None,
        epilog: str | None = None,
        exist_ok: bool = False,
    ) -> Callable[[Type[Subclass]], Type[Subclass]]:
        registry = Subcommand._registry[cls]

        def wrapper(subclass: Type[Subclass]) -> Type[Subclass]:
            info = SubcommandInfo(
                name=name or cls.camel_to_snake(subclass.__name__),
                usage=usage,
                description=description or subclass.__doc__,
                epilog=epilog,
            )
            subclass._cmd_info = info

            if not exist_ok and name in registry:
                raise ValueError(f"Subcommand '{name}' was already registered.")

            registry[info.name] = subclass
            return subclass

        return wrapper

    @classmethod
    def get_info(cls) -> SubcommandInfo:
        if not hasattr(cls, "_cmd_info"):
            cls._cmd_info = SubcommandInfo(
                name=cls.camel_to_snake(cls.__name__),
                description=cls.__doc__,
            )
        return cls._cmd_info

    @staticmethod
    def camel_to_snake(text: str) -> str:
        underscored = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", text)
        return re.sub("([a-z0-9])([A-Z])", r"\1_\2", underscored).lower()

    def __init__(
        self,
        parser_or_subparsers: argparse.ArgumentParser | argparse._SubParsersAction | None = None,
        subcommand_info: SubcommandInfo | None = None,
    ) -> None:
        cls = type(self)
        info = subcommand_info or self.get_info()

        if isinstance(parser_or_subparsers, argparse.ArgumentParser):
            self._parser = parser_or_subparsers
        elif isinstance(parser_or_subparsers, argparse._SubParsersAction):
            self._parser = parser_or_subparsers.add_parser(
                name=info.name,
                usage=info.usage,
                description=info.description,
                epilog=info.epilog,
            )
        else:
            self._parser = argparse.ArgumentParser()

        self.setup()
        self._parser.set_defaults(**{self._func_key: self.run})

        registry = Subcommand._registry[cls]
        if registry:
            subparsers = self.parser.add_subparsers()
            for subclass in registry.values():
                subclass(subparsers)

    def __call__(self, args: argparse.Namespace | None = None) -> None:
        if not args:
            args = self.parser.parse_args()

        func = cast(
            Optional[Callable[[argparse.Namespace], None]],
            getattr(args, self._func_key, None),
        )
        if func:
            func(args)
        else:
            self.parser.print_help()

    @property
    def parser(self) -> argparse.ArgumentParser:
        return self._parser

    def setup(self) -> None:
        """setup parser"""

    def run(self, args: argparse.Namespace) -> None:
        self.parser.print_help()



from __future__ import annotations

import datetime
import inspect
from pathlib import Path
from typing import Any, Callable, NamedTuple

from cachestore.hashers import Hasher


class FunctionInfo(NamedTuple):
    name: str
    filename: Path
    source: str

    @classmethod
    def build(cls, func: Callable[..., Any]) -> "FunctionInfo":
        if hasattr(func, "__cachesore_funcinfo"):
            funcinfo = getattr(func, "__cachesore_funcinfo")
            assert isinstance(funcinfo, FunctionInfo)
            return funcinfo
        filename = Path(inspect.getabsfile(func))
        modulename = inspect.getmodulename(str(filename)) or ""
        name = f"{modulename}.{func.__qualname__}"
        lines, _ = inspect.getsourcelines(func)
        source = "".join(lines)
        return cls(name, filename, source)

    def hash(self, hasher: Hasher) -> str:
        return hasher(self)

    def to_dict(self) -> dict[str, Any]:
        return {
            "name": self.name,
            "filename": str(self.filename),
            "source": self.source,
        }

    @classmethod
    def from_dict(cls, d: dict[str, Any]) -> "FunctionInfo":
        return cls(
            name=d["name"],
            filename=Path(d["filename"]),
            source=d["source"],
        )


class ExecutionInfo(NamedTuple):
    params: dict[str, Any]

    @classmethod
    def build(
        cls,
        func: Callable[..., Any],
        *args: Any,
        **kwargs: Any,
    ) -> "ExecutionInfo":
        signature = inspect.signature(func)
        arguments: list[tuple[str | None, Any]] = [(None, v) for v in args]
        arguments += list(kwargs.items())

        position = 0
        params: dict[str, Any] = {}
        try:
            for key, param in signature.parameters.items():
                if position >= len(arguments):
                    if param.default != inspect.Parameter.empty:
                        params[key] = param.default
                    elif param.kind == inspect.Parameter.VAR_POSITIONAL:
                        params[key] = []
                    elif param.kind == inspect.Parameter.VAR_KEYWORD:
                        params[key] = {}
                    else:
                        raise AssertionError("There are some missing arguments.")
                elif param.kind == inspect.Parameter.POSITIONAL_ONLY:
                    if arguments[position][0] is None and param.default == inspect.Parameter.empty:
                        params[key] = arguments[position][1]
                        position += 1
                elif param.kind == inspect.Parameter.POSITIONAL_OR_KEYWORD:
                    if key == arguments[position][0] or arguments[position][0] is None:
                        params[key] = arguments[position][1]
                        position += 1
                    elif param.default != inspect.Parameter.empty:
                        params[key] = param.default
                elif param.kind == inspect.Parameter.VAR_POSITIONAL:
                    params[key] = []
                    while (
                        position < len(arguments)
                        and arguments[position][0] is None
                        and arguments[position][0] not in signature.parameters
                    ):
                        params[key].append(arguments[position][1])
                        position += 1
                elif param.kind == inspect.Parameter.KEYWORD_ONLY:
                    if key == arguments[position][0]:
                        params[key] = arguments[position][1]
                        position += 1
                    elif param.default != inspect.Parameter.empty:
                        params[key] = param.default
                elif param.kind == inspect.Parameter.VAR_KEYWORD:
                    if key not in params:
                        params[key] = {}
                    while position < len(arguments) and arguments[position][0] not in signature.parameters:
                        params[key][arguments[position][0]] = arguments[position][1]
                        position += 1
                    params[key] = dict(sorted(params[key].items()))
                else:
                    raise AssertionError("This statement will be never executed.")

            for extrakey, value in arguments[position:]:
                assert extrakey is not None, extrakey
                assert extrakey not in params, extrakey
                params[extrakey] = value

            assert list(params.keys()) == list(signature.parameters.keys()), f"{params=}, {signature.parameters=}"
        except AssertionError as err:
            raise ValueError(
                f"Invalid arguments of {func.__module__}.{func.__name__}:\n\t"
                f"Signature : {signature}\n\t"
                f"Given args: {args=}, {kwargs=}"
            ) from err

        return cls(params)

    def hash(self, hasher: Hasher) -> str:
        return hasher(self)


class CacheInfo(NamedTuple):
    function: FunctionInfo
    parameters: dict[str, Any]
    executed_at: datetime.datetime
    expired_at: datetime.datetime | None

    def to_dict(self) -> dict[str, Any]:
        return {
            "function": self.function.to_dict(),
            "parameters": {k: repr(v) for k, v in self.parameters.items()},
            "executed_at": self.executed_at.isoformat(),
            "expired_at": self.expired_at.isoformat() if self.expired_at else None,
        }

    @classmethod
    def from_dict(cls, d: dict[str, Any]) -> "CacheInfo":
        return cls(
            function=FunctionInfo.from_dict(d["function"]),
            parameters=d["parameters"],
            executed_at=datetime.datetime.fromisoformat(d["executed_at"]),
            expired_at=datetime.datetime.fromisoformat(d["expired_at"]) if d["expired_at"] else None,
        )



import argparse
import sys

from cachestore.cache import Cache
from cachestore.commands.subcommand import Subcommand
from cachestore.util import import_modules, safe_import_object


@Subcommand.register("prune")
class PruneCommand(Subcommand):
    """prune unreferenced caches"""

    def setup(self) -> None:
        self.parser.add_argument("cache", help="cache name")
        self.parser.add_argument(
            "--include-package",
            action="append",
            default=[],
            help="additinoal packages to include",
        )

    def run(self, args: argparse.Namespace) -> None:
        if args.include_package:
            import_modules(args.include_package)

        cache = Cache.by_name(args.cache)
        if cache is None:
            cache = safe_import_object(args.cache)

        if cache is None:
            print(f"Given cache name is not found: {args.cache}", file=sys.stderr)
            sys.exit(1)

        cache.prune()



from __future__ import annotations

import abc
from configparser import SectionProxy
from typing import Any, Type, TypeVar

Self = TypeVar("Self", bound="Hasher")


class Hasher(abc.ABC):
    @abc.abstractmethod
    def __call__(self, obj: Any) -> str:
        raise NotImplementedError

    @classmethod
    def from_config(cls: Type[Self], config: SectionProxy) -> Self:
        raise NotImplementedError



from __future__ import annotations

import datetime
import inspect
import json
from functools import wraps
from logging import getLogger
from typing import Any, Callable, Iterator, TypeVar, cast

from cachestore.config import CacheSettings, Config
from cachestore.formatters import Formatter
from cachestore.hashers import Hasher
from cachestore.metadata import CacheInfo, ExecutionInfo, FunctionInfo
from cachestore.storages import Storage
from cachestore.util import find_variable_path

logger = getLogger(__name__)

T = TypeVar("T")


class Cache:
    _cache_registry: list["Cache"] = []

    def __init__(
        self,
        name: str | None = None,
        *,
        storage: Storage | None = None,
        formatter: Formatter | None = None,
        hasher: Hasher | None = None,
        disable: bool | None = None,
        config: Config | None = None,
    ) -> None:
        self.config = config or Config()

        self._name = name
        self._storage = storage
        self._formatter = formatter
        self._hasher = hasher
        self._disable = disable

        self._settings: CacheSettings | None = None
        self._function_registry: dict[str, FunctionInfo] = {}

        self._cache_registry.append(self)

    @classmethod
    def by_name(self, name: str) -> Cache | None:
        for cache in self._cache_registry:
            if cache.name == name:
                return cache
        return None

    @property
    def name(self) -> str:
        if self._name is None:
            self._name = find_variable_path(self)
        if self._name is None:
            raise RuntimeError("Cannot get cache name.")
        return self._name

    @property
    def settings(self) -> CacheSettings:
        if self._settings is None:
            self._settings = self.config.cache_settings(self.name)
            if self._storage is not None:
                self._settings.storage = self._storage
            if self._formatter is not None:
                self._settings.formatter = self._formatter
            if self._hasher is not None:
                self._settings.hasher = self._hasher
            if self._disable is not None:
                self._settings.disable = self._disable
        return self._settings

    @property
    def storage(self) -> Storage:
        return self.settings.storage

    @property
    def formatter(self) -> Formatter:
        return self.settings.formatter

    @property
    def hasher(self) -> Hasher:
        return self.settings.hasher

    @property
    def disable(self) -> bool:
        return self.settings.disable

    def _get_key(self, funcinfo: FunctionInfo, execinfo: ExecutionInfo) -> str:
        return ".".join((funcinfo.hash(self.hasher), execinfo.hash(self.hasher)))

    def _get_metakey(self, key: str) -> str:
        return f"metadata-{key}"

    def _is_metakey(self, key: str) -> bool:
        return key.startswith("metadata-")

    def __call__(
        self,
        *,
        ignore: set[str] | None = None,
        expire: int | datetime.timedelta | datetime.date | datetime.datetime | None = None,
        formatter: Formatter | None = None,
        disable: bool | None = None,
    ) -> Callable[[Callable[..., T]], Callable[..., T]]:
        def decorator(func: Callable[..., T]) -> Callable[..., T]:
            funcinfo = FunctionInfo.build(func)
            self._function_registry[funcinfo.hash(self.hasher)] = funcinfo

            function_settings = self.config.function_settings(f"{self.name} {funcinfo.name}")
            if ignore is not None:
                function_settings.ignore = ignore
            if expire is not None:
                function_settings.expire = expire
            if disable is not None:
                function_settings.disable = disable
            if formatter is not None:
                function_settings.formatter = formatter

            @wraps(func)
            def wrapper(*args: Any, **kwargs: Any) -> T:
                disable = self.disable if function_settings.disable is None else function_settings.disable
                ignore = function_settings.ignore
                expired_at = function_settings.expired_at
                executed_at = datetime.datetime.now()

                storage = self.storage
                formatter = function_settings.formatter or self.formatter

                if disable:
                    logger.info("[%s] Disable cache.", funcinfo.name)
                    return func(*args, **kwargs)

                execinfo = ExecutionInfo.build(func, *args, **kwargs)
                for paramname in ignore:
                    del execinfo.params[paramname]

                key = self._get_key(funcinfo, execinfo)
                metakey = self._get_metakey(key)

                if storage.exists(metakey):
                    with storage.open(metakey, "rt") as file:
                        cacheinfo = CacheInfo.from_dict(json.load(file))
                    if cacheinfo.expired_at is not None and cacheinfo.expired_at <= executed_at:
                        logger.info("[%s] Cache was expired, so remove existing artifact.", funcinfo.name)
                        storage.remove(key)
                        storage.remove(metakey)

                if storage.exists(key) and (expired_at is None or expired_at > executed_at):
                    logger.info("[%s] Cache exists", funcinfo.name)
                    with self.storage.open(key, formatter.READ_MODE) as file:
                        artifact = cast(T, formatter.read(file))
                else:
                    logger.info("[%s] Cache does not exists.", funcinfo.name)
                    artifact = func(*args, **kwargs)

                    logger.info("[%s] Store new artifact.", funcinfo.name)
                    with storage.open(key, formatter.WRITE_MODE) as file:
                        formatter.write(file, artifact)

                    logger.info("[%s] Export metadata.", funcinfo.name)
                    cacheinfo = CacheInfo(
                        function=funcinfo,
                        parameters=execinfo.params,
                        expired_at=expired_at,
                        executed_at=executed_at,
                    )
                    with storage.open(metakey, "wt") as file:
                        json.dump(cacheinfo.to_dict(), file)

                return artifact

            setattr(wrapper, "__signature__", inspect.signature(func))
            setattr(wrapper, "__annotations__", func.__annotations__)
            setattr(wrapper, "__cachesore_funcinfo", funcinfo)

            return wrapper

        return decorator

    def exists(self, func: Callable[..., Any] | FunctionInfo) -> bool:
        current = datetime.datetime.now()
        exists = False
        for cacheinfo in self.info(func):
            exists |= cacheinfo.expired_at is None or cacheinfo.expired_at > current
        return exists

    def remove(self, func: Callable[..., Any] | FunctionInfo) -> None:
        if not isinstance(func, FunctionInfo):
            func = FunctionInfo.build(func)
        prefix = func.hash(self.hasher)
        for key in self.storage.filter(prefix=prefix):
            self.storage.remove(key)

    def funcinfos(self) -> list[FunctionInfo]:
        return list(self._function_registry.values())

    def info(self, func: Callable[..., Any] | FunctionInfo) -> Iterator[CacheInfo]:
        if not isinstance(func, FunctionInfo):
            func = FunctionInfo.build(func)
        prefix = func.hash(self.hasher)
        for key in self.storage.filter(prefix=prefix):
            metakey = self._get_metakey(key)
            with self.storage.open(metakey, "rt") as file:
                yield CacheInfo.from_dict(json.load(file))

    def prune(self) -> None:
        for key in self.storage.all():
            if not any(
                key.startswith(funchash) or key.startswith(self._get_metakey(funchash))
                for funchash in self._function_registry
            ):
                logger.info("remove %s", key)
                self.storage.remove(key)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['cachestore',
 'cachestore.commands',
 'cachestore.common',
 'cachestore.formatters',
 'cachestore.hashers',
 'cachestore.storages']

package_data = \
{'': ['*']}

entry_points = \
{'console_scripts': ['cachestore = cachestore.__main__:run']}

setup_kwargs = {
    'name': 'cachestore',
    'version': '0.2.0',
    'description': 'Function Cache Management Tool for Python',
    'long_description': '# CacheStore\n\n[![Actions Status](https://github.com/altescy/cachestore/workflows/CI/badge.svg)](https://github.com/altescy/cachestore/actions/workflows/ci.yml)\n[![Python version](https://img.shields.io/pypi/pyversions/cachestore)](https://github.com/altescy/cachestore)\n[![License](https://img.shields.io/github/license/altescy/cachestore)](https://github.com/altescy/cachestore/blob/master/LICENSE)\n[![pypi version](https://img.shields.io/pypi/v/cachestore)](https://pypi.org/project/cachestore/)\n\n**CacheStore** is a simple cache management system for Python functions.\nYou can reuse the cached results even accross different executions.\n\n**cachestore** command enables you to manage the cached results from command line.\nPlease see `--help` for more details.\n\n**Features**\n\n- Caching execution results by decorating target functions easily\n- Exporting caches into an external storage to reuse them access different exeutions\n- Detecting appropreate caches based on argumetns/source code of functions\n- Changing cache behavior via configuration file (see [exmaples](./examples))\n- Providing a useful command line tool to manage caches\n- Written in pure Python, no external dependencies\n\n## Installation\n\n```bash\npip install cachestore\n```\n\n## Usage\n\n### Python\n\n```python\nfrom cachestore import Cache\n\ncache = Cache()\n\n@cache()\ndef awesome_function(x, *, y="y", **kwargs):\n    ...\n```\n\n### CLI\n\n```bash\n$ cachestore --help\nusage: cachestore\n\npositional arguments:\n  {list,prune,remove}\n\noptional arguments:\n  -h, --help           show this help message and exit\n  --version            show program\'s version number and exit\n```\n',
    'author': 'altescy',
    'author_email': 'altescy@fastmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/altescy/cachestore',
    'packages': packages,
    'package_data': package_data,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



import argparse
import datetime
import sys
from pathlib import Path

from cachestore.cache import Cache
from cachestore.commands.subcommand import Subcommand
from cachestore.common import Selector, Table
from cachestore.util import import_modules, safe_import_object


@Subcommand.register("list")
class ListCommand(Subcommand):
    """list functions and cache statuses"""

    def setup(self) -> None:
        self.parser.add_argument(
            "cache",
            help="cache name",
        )
        self.parser.add_argument(
            "--include-package",
            action="append",
            default=[],
            help="additinoal packages to include",
        )

    def run(self, args: argparse.Namespace) -> None:
        if args.include_package:
            import_modules(args.include_package)

        cache = Cache.by_name(args.cache)
        if cache is None:
            cache = safe_import_object(args.cache)

        if cache is None:
            print(f"Given cache name is not found: {args.cache}", file=sys.stderr)
            sys.exit(1)

        table = Table(columns=["name", "function", "filename", "cache", "last_executed_at"])
        for funcinfo in cache.funcinfos():
            cacheinfos = list(cache.info(funcinfo))
            table.add(
                {
                    "name": args.cache,
                    "function": funcinfo.name,
                    "filename": str(funcinfo.filename.relative_to(Path.cwd())),
                    "cache": "\033[32m✓\033[39m" if cache.exists(funcinfo) else "",
                    "last_executed_at": max(info.executed_at for info in cacheinfos).strftime("%Y-%m-%d %H:%M:%S")
                    if cacheinfos
                    else "",
                }
            )

        table.sort("last_executed_at", desc=True)
        table.show()


@ListCommand.register("details")
class DetailsCommand(ListCommand):
    """list cache status details of specified function"""

    def setup(self) -> None:
        self.parser.add_argument(
            "-f",
            "--function",
            default=None,
            help="function name to show details",
        )

    def run(self, args: argparse.Namespace) -> None:
        if args.include_package:
            import_modules(args.include_package)

        table = Table(columns=["cache", "function", "filename", "params", "executed_at", "expired_at"])

        cache = safe_import_object(args.cache)
        assert isinstance(cache, Cache)

        funcinfos = {info.name: info for info in cache.funcinfos()}

        if args.function:
            funcname = args.function
        else:
            funcname = Selector()(list(funcinfos))

        if not funcname or funcname not in funcinfos:
            print("No information to show.")
            return

        funcinfo = funcinfos[funcname]

        for cacheinfo in cache.info(funcinfo):
            table.add(
                {
                    "cache": args.cache,
                    "function": funcinfo.name,
                    "filename": str(funcinfo.filename.relative_to(Path.cwd())),
                    "params": ", ".join(f"{k}={v}" for k, v in cacheinfo.parameters.items()),
                    "executed_at": cacheinfo.executed_at.strftime("%Y-%m-%d %H:%M:%S"),
                    "expired_at": (
                        cacheinfo.expired_at.strftime("%Y-%m-%d %H:%M:%S")
                        if cacheinfo.expired_at > datetime.datetime.now()
                        else f"\033[31m{cacheinfo.expired_at.strftime('%Y-%m-%d %H:%M:%S')}\033[39m"
                    )
                    if cacheinfo.expired_at
                    else "",
                }
            )

        table.sort("executed_at", desc=True)
        table.show()



import abc
from configparser import SectionProxy
from typing import IO, Any, ClassVar, Type, TypeVar

Self = TypeVar("Self", bound="Formatter")


class Formatter(abc.ABC):
    READ_MODE: ClassVar[str]
    WRITE_MODE: ClassVar[str]

    @abc.abstractmethod
    def write(self, file: IO[Any], obj: Any) -> None:
        raise NotImplementedError

    @abc.abstractmethod
    def read(self, file: IO[Any]) -> Any:
        raise NotImplementedError

    @classmethod
    def from_config(cls: Type[Self], config: SectionProxy) -> Self:
        raise NotImplementedError



import pickle
from configparser import SectionProxy
from typing import IO, Any, ClassVar, Type, TypeVar

from cachestore.formatters.formatter import Formatter

Self = TypeVar("Self", bound="PickleFormatter")


class PickleFormatter(Formatter):
    READ_MODE: ClassVar = "rb"
    WRITE_MODE: ClassVar = "wb"

    def write(self, file: IO[Any], obj: Any) -> None:
        pickle.dump(obj, file)

    def read(self, file: IO[Any]) -> Any:
        return pickle.load(file)

    @classmethod
    def from_config(cls: Type[Self], config: SectionProxy) -> Self:
        return cls()
